/*
 * Copyright 1996, Be Inc.
 *
 * Adapted from code originated from:
 *   Thomas Davis
 */



#include <OS.h>
#include <Drivers.h>
#include <KernelExport.h>
#include <stdlib.h>

#include "qcam_driver.h"

/*
	This quickcam struct is used internally in the driver to 
	support all aspects of talking to the camera.
 */

typedef struct quickcam {
	int pixelsperline;
	int	pixelsperframe;
	vuchar	*iobase;
	vuchar	*datareg;
	vuchar	*controlreg;
	vuchar	*statusreg;
	unsigned int port;			/* port address */
	unsigned char *buffer;		/* video buffer */
	unsigned char bpp;			/* bits-per-pixel */
	unsigned char mode;			/* transfer mode */
	unsigned char status;		/* port status (uni/bi) */
	unsigned char videoparam;	/* videoparam */
	unsigned char brightness;	/* brightness */
	unsigned char blacklevel;	/* whitebalance */
	unsigned char contrast;		/* contrast */
	unsigned char transferfactor;
	unsigned char transfersperline;
	char init;			/* send the values to the camera */
}qcam_data;

	
// Some function prototypes
void udelay(double microseconds);

#define printk dprintf

int quickcam_debug = 0;

#define HIGHNYBBLE(x) (((x&0xf0)>>4))
#define LOWNYBBLE(x) ((x&0x0f))

#define PCACKRAISE	*qc->controlreg &= PCAckRaise;
#define PCACKLOWER	*qc->controlreg &= PCAckLower;

#define ISCAMREADY (*qc->statusreg&CamRdy1)



int topline = 0x01;
int leftcolumn = 0x07;	/* leftcolumn/2 */

struct quickcam camera, *qc;

void
udelay(double microseconds)
{
	spin(microseconds);
}

static void 
reset_quickcam(struct quickcam *cam)
{
	
	if (cam->status == QC_BI_TEST_BYTE)
		cam->mode = QC_NYBBLE_TRANSFER;
		
	if (quickcam_debug > 1) 
	{
		printk("reset_quickcam - status = 0x%02x \n", cam->status);
	}
	
	// To reset the camera, pull Reset_N low for a little bit
	*cam->controlreg = 0x0b;
	udelay(25);

	// Now pull it back
	*cam->controlreg = 0x0e;
	udelay(25);
}

static int 
getstatus()
{
	unsigned char status1, status2;

	*qc->controlreg = 0x06;

	do {
		status1 = *qc->statusreg;
	} while (!(status1 & 0x08)) ;

	*qc->controlreg = 0x0e;

	do {
		status2 = *qc->statusreg;
	} while (status2 & 0x08) ;

	if (quickcam_debug > 2) 
	{
		printk("status1 = 0x%02x ", status1);
		printk("status2 = 0x%02x ", status2);
		printk("status = 0x%02x\n",((status1 & 0xF0) | ( status2 >> 4 )));
	}

	return ((status1 & 0xF0 ) | ( status2 >> 4 ));
}


static int 
sendbyte(uchar value)
{
	*qc->datareg = value;
	*qc->datareg = value;
	*qc->datareg = value;

	return (getstatus());
}

static int 
sendcmd(uchar cmd, uchar value)
{
	if (quickcam_debug > 2)
		printk("cmd = 0x%02x, value = 0x%02x\n", cmd, value);

	if (sendbyte(cmd) != cmd )
		return(1);

	if (sendbyte(value) != value )
		return(1);

	return 0;
}

static void 
set_videoparam(struct quickcam *qc)
{
	/* start from scratch */
	qc->videoparam = 0;
	
	qc->videoparam |= qc->mode;
	qc->videoparam |= qc->bpp;
	qc->videoparam |= qc->transferfactor;
	
	if (quickcam_debug > 1)
		printk("setting videoparam to %d\n", qc->videoparam);

}


static void 
qcam_init(struct quickcam *qc)
{
	int transline = 0;
	int y=240;
	
	set_videoparam(qc);
	qc->pixelsperframe = 320*240;
	
	// Setup some other parameters
	// Figure out how many transfers per line
	transline = 320;
	// take into account the pixel depth
	if (qc->bpp == QC_6BIT_PIXEL)
		transline /= 4;
	else 
		transline /= 2;
			
	// Account for compression factor
	if (qc->transferfactor == QC_FACTOR_21) {
		transline /= 2;
		y /= 2;
		qc->pixelsperframe /= 4;
	} else if (qc->transferfactor == QC_FACTOR_41) {
		transline /= 4;
		y /= 4;
		qc->pixelsperframe /= 16;
	};
	
	// Assign whatever is left
	qc->transfersperline = transline;
	
	sendcmd(QC_SETEXPOSURE, qc->brightness);
	sendcmd(QC_SETEXPOSURE, 0x01);
	sendcmd(QC_SETEXPOSURE, 0x01);
	sendcmd(QC_SETEXPOSURE, qc->brightness);
	sendcmd(QC_SETEXPOSURE, qc->brightness);

	sendcmd(QC_SETTOP, topline);
	sendcmd(QC_SETLEFT, leftcolumn);
	sendcmd(QC_SETEXPOSURE, qc->brightness);
	sendcmd(QC_SETNUMH, qc->transfersperline);
	sendcmd(QC_SETNUMV, y);
	sendcmd(QC_SETCONTRAST, qc->contrast);
	sendcmd(QC_SETBLACKLEVEL, qc->blacklevel);

	qc->init = 0;
}

static int 
uni_4bit(struct quickcam *qc, unsigned char *buffer)
{
	unsigned char word1=0;
	unsigned char *tmp = buffer;
	unsigned char *end;
	
	end = buffer+qc->pixelsperframe;
	while(tmp < end)
	{
		*qc->controlreg = 0x06;
		do {
			word1 = *qc->statusreg;
		} while (!(word1 & 0x08));
		*tmp++ = ((16-(word1 >> 4)) & 0x0F) << 1;

		*qc->controlreg = 0x0e;
		do {
			word1 = *qc->statusreg;
		} while (word1 & 0x08);
		*tmp++ = ((16-(word1 >> 4)) & 0x0F) << 1;
	}

	return (1);

}

static int 
uni_6bit(struct quickcam *qc, unsigned char *buffer)
{
	unsigned char byte1, byte2, byte3;
	unsigned char *tmp = buffer;
	int counter=0;
	int x, y;
	
	while (tmp < buffer+qc->pixelsperframe)
	{
		*qc->controlreg = 0x06;
		do {
			byte1 = *qc->statusreg;
		} while (!(byte1 & 0x08));

		*qc->controlreg = 0x0e;
		do {
			byte2 = *qc->statusreg;
		} while (byte2 & 0x08);
		*tmp++ = (63 - (((byte1 & 0xF0) >> 2) | (byte2 >> 6)))>>1;		

		*qc->controlreg = 0x06;
		do {
			byte1 = *qc->statusreg;
		} while (!(byte1 & 0x08));
		*tmp++ = (63 - ((byte2 & 0x30) | (byte1 >> 4)))>>1;

		*qc->controlreg = 0x0e;
		do {
			byte1 = *qc->statusreg;
		} while (byte1 & 0x08);

		*qc->controlreg = 0x06;
		do {
			byte2 = *qc->statusreg;
		} while (!(byte2 & 0x08));
		*tmp++ = (63 - (((byte1 & 0xF0) >> 2) | (byte2 >> 6)))>>1;		

		*qc->controlreg = 0x0e;
		do {
			byte1 = *qc->statusreg;
		} while (byte1 & 0x08);
		*tmp++ = (63 - ((byte2 & 0x30) | (byte1 >> 4)))>>1;
	}

	return (1);
}

static int 
waitfor_bi()
{
	unsigned char status1 = 0, status2 = 1;

	// perform port turn around handshake
	// Raise PCAck to initiate handshake

	*qc->controlreg = 0x26;
	*qc->controlreg = 0x26;

	if (quickcam_debug > 1) 
		printk("waiting for camera to say \"Go!\"\n");

	do {
		status1 = *qc->statusreg;
	} while (!(status1 & 0x08));

	// acknowledge by lowering PCAck
	*qc->controlreg = 0x2f;

	do {
		status2 = *qc->statusreg;
	} while (status2 & 0x08);

	if (quickcam_debug > 1) 
	{
		printk("waitfor_bi - status1 = 0x%02x ", status1);
		printk("waitfor_bi - status2 = 0x%02x ", status2);
		printk("waitfor_bi - status = 0x%02x\n",((status1 & 0xF0) | ( status2 >> 4 )));
	}

	return ((status1 & 0xF0 ) | ( status2 >> 4 ));

}

static void 
bi_4bit(struct quickcam *qc, unsigned char *buffer)
{
	strcpy((char*)buffer,"bi_4bit");
	printk("not implemented at this time.\n");
}

static void 
bi_6bit(struct quickcam *qc, unsigned char *buffer)
{
	unsigned short word1=66;
	unsigned char *tmp = buffer;
	unsigned char byte1, byte2;
	unsigned char ready;
	unsigned char *end;
	
	// Wait for camera to come ready
	waitfor_bi();

	end = buffer+qc->pixelsperframe;

	// Now start the actual frame transfer
	while (tmp < end)
	{
		// Raise PCAck to begin transfer of first nybble
		*qc->controlreg = 0x26;
		
		// keep reading the ports until CamRdy2 (datareg bit 0)
		// is a high value.  When this occurs, the camera is
		// indicating that the first nybble is ready
		do {
			byte1 = *qc->datareg;
			byte2 = *qc->statusreg;
		} while (!(byte1 & 1));
		
		*tmp++ = 63-((byte2 >> 3) | ((byte1 << 4) & 0x20));
		*tmp++ = 63-(byte1 >> 2);
		
		// lower PCAck to indicate readiness for next nybble
		*qc->controlreg = 0x2f;

		// keep reading the ports until CamRdy2 (datareg bit 0)
		// is a low value.  When this occurs, the camera is
		// indicating that the second nybble is ready
		do {
			byte1 = *qc->datareg;
			byte2 = *qc->statusreg;
		} while ((byte1 & 1));

		*tmp++ = 63-((byte2 >> 3) | ((byte1 << 4) & 0x20));
		*tmp++ = 63-(byte1 >> 2);
	}
}

static void 
doscan(struct quickcam *qc, unsigned char * buffer)
{

	/*
		Make sure framing parameters are correct
		before transfer starts
	*/
	sendcmd(QC_SETEXPOSURE, qc->brightness);
	
	/*
		Tell camera to start sending video
	*/
	sendcmd(QC_SENDVIDEO, qc->videoparam);

	switch (qc->bpp) 
	{
		case QC_4BIT_PIXEL:
			if (qc->mode == QC_BYTE_TRANSFER && qc->status != QC_BI_TEST_BYTE)
			{
				bi_4bit(qc, buffer);
			} else
			{
				uni_4bit(qc, buffer);
			}
		break;
		
		case QC_6BIT_PIXEL:
			if (qc->mode == QC_BYTE_TRANSFER && qc->status != QC_BI_TEST_BYTE) 
			{
				bi_6bit(qc, buffer);
			} else
			{
				uni_6bit(qc, buffer);
			}
		break;
	}

	/*
		End of frame
		1) send a pulse by raising voltalge of PCAck (inverted)
		2) Keep PCAck high for not less than 3 microseconds
		3) Lower PCAck to indicate that the camera should be back
			in command mode.
	*/

	*qc->controlreg = 0x0f;
	udelay(10);
	*qc->controlreg = 0x0b;
	

	if (quickcam_debug > 0)
		printk("got scan!\n");
}


static long 
qcam_open(device_info *dev, ulong flags)
{
	dprintf("qcam_open - BEGIN\n");


	if (quickcam_debug > 0) {
		printk("\nquickcam open: dev flags=%d\n", flags);
		printk("reseting camera...\n");
	}

	reset_quickcam(qc);

	if (quickcam_debug > 0) {
		printk("initing camera...\n");
	}

	qcam_init(qc);

	if (quickcam_debug > 0) 
	{
		printk("scanning in %d bpp mode...\n", qc->bpp);
	}

	if (quickcam_debug > 0)
		printk("quickcam open done\n" );

	return B_NO_ERROR;
}

static long 
qcam_close(device_info *dev)
{
	if (quickcam_debug > 0)
		printk("quickcam close OK\n" );

	return B_NO_ERROR;
}

static long 
qcam_read(device_info *dev, void *buf, ulong count, ulong pos)
{
	if (qc->init == 1)
		qcam_init(qc);

	if (count >= qc->pixelsperframe) {
		doscan(qc, buf);
		return qc->pixelsperframe;
	};

	return 0;
}

static long 
qcam_write(device_info *dev, void *buf, ulong count, ulong pos)
{
	if (quickcam_debug > 1)
		printk("\nquickcam write: try write %d bytes\n", count);

	return 0;
}

static long 
qcam_control(device_info *dev, ulong cmd, void *arg)
{
	int retval = B_NO_ERROR;
	struct qcam *camera = arg;

	if (quickcam_debug > 1)
		printk("\nquickcam ioctl: try %d\n",cmd);

	switch (cmd) 
	{
		case QC_GET:
			camera->bpp = qc->bpp;
			camera->mode = qc->mode;
			camera->brightness = qc->brightness;
			camera->contrast = qc->contrast;
			camera->status = qc->status;
			camera->factor = qc->transferfactor;
			camera->whitebalance = qc->blacklevel;
		break;

		case QC_SET:
			qc->bpp = camera->bpp;
			qc->mode = camera->mode;
			qc->brightness = camera->brightness;
			qc->contrast = camera->contrast;
			qc->status = camera->status;
			qc->transferfactor = camera->factor;
			qc->blacklevel = camera->whitebalance;
			qc->init = 1;
		break;

		default:
			retval = B_ERROR;
	}
	
	return retval;
}

long 
init_driver(void) 
{
	vuchar		*base;
	int ret = B_NO_ERROR;

	set_dprintf_enabled(1);
	printk("quickcam driver compiled: %s %s\n", __DATE__, __TIME__);

	qc = (struct quickcam *) &camera;

	base			= (vuchar *) isa_address (QC_IO_BASE);
	qc->port		= (int)base;
	qc->iobase 		= base;
	qc->datareg 	= base + QC_DATA_REGISTER;
	qc->statusreg   = base + QC_STATUS_REGISTER;
	qc->controlreg  = base + QC_CONTROL_REGISTER;

	if (quickcam_debug > 3)
		printk("quickcam: init driver \n");

	printk("quickcam: init port=0x%x debug=%d\n", qc->iobase, quickcam_debug);
   

	qc->contrast = 104;
	qc->brightness = 125;
	qc->blacklevel = 100;
	qc->videoparam = 0;
	qc->mode = QC_NYBBLE_TRANSFER;
	qc->bpp = QC_6BIT_PIXEL;
	qc->transferfactor = QC_FACTOR_41;
	qc->init = 1;

	// For bidirectional ports, if you set bit 5
	// of the control port, then the data pins will
	// be set for input.
	
	// Write some bits out the data port.  If they come
	// back the same as they went in, then the port is
	// not set for bidirectional operation, or it is not
	// a bidirectional port.
	// Set port for data output

	// But alas, this has problems, so try it yourself some time.
/*	
	*qc->controlreg = 0;
	udelay(25);
	// write out some data
	*qc->datareg = QC_BI_TEST_BYTE;
	
	udelay(25);
	

	// set port for bidirectional input
	*qc->controlreg = 0x20;
	udelay(25);
	// read in some data
	qc->status = *qc->datareg;

	// The status byte is the value read back from the data
	// port after it was supposed to be in bidirectional mode.
	// If this is the same as the value written out, then the 
	// port is not bidirectional, otherwise, it probably is.
	if (qc->status != QC_BI_TEST_BYTE) 
	{
		printk("qcam on a bi-directional port..\n");
	} else
		printk("qcam apparently on a uni-directional port..\n");
*/

	qc->status = QC_BI_TEST_BYTE;

	// Now reset the camera
	reset_quickcam(qc);

	printk("quickcam device driver loaded\n");

	return B_NO_ERROR;

}


device_entry	devices[] = {
	{
		"/dev/quickcam",	// name
		qcam_open,			// -> open entry point 
		qcam_close, 		// -> close entry point 
		qcam_control,		// -> control entry point 
		qcam_read,			// -> read entry point 
		qcam_write			// -> write entry point 
	},
	0
};


